// CycmNative.cpp
// (c) 2011 Egtra

#include "stdafx.h"
#include <vector>
#include <boost/scoped_array.hpp>
#include <boost/interprocess/smart_ptr/unique_ptr.hpp>
#include <boost/interprocess/streams/vectorstream.hpp>
#include <boost/implicit_cast.hpp>

#include "libtta.h"
#include "config.h"

#include "common.h"

struct SafeArrayDeleter
{
	void operator()(SAFEARRAY* psa) const
	{
		SafeArrayDestroy(psa);
	}
};

typedef boost::interprocess::unique_ptr<SAFEARRAY, SafeArrayDeleter> SafeArrayPtr;

//struct SafeArrayAccessDeleter
//{
//	explicit SafeArrayAccessDeleter(SAFEARRAY* psa = NULL) : psa(psa) {}
//	SafeArrayAccessDeleter(SafeArrayAccessDeleter const& y) : psa(y.psa) {}
//	SafeArrayAccessDeleter& operator =(SafeArrayAccessDeleter const& y)
//	{
//		psa = y.psa;
//		return *this;
//	}
//	void operator ()(void*) const
//	{
//		if (psa != NULL)
//		{
//			SafeArrayUnaccessData(psa);
//		}
//	}
//private:
//	SAFEARRAY* psa;
//};
//
//// template aliasgtemplate<typename T> using SafeArrayAccessPtr<T> = boost::interprocess::unique_ptr<T, SafeArrayAccessDeleter>Ə
//typedef boost::interprocess::unique_ptr<BYTE, SafeArrayAccessDeleter> SafeArrayAccessBytePtr;
//
//SafeArrayAccessBytePtr SafeArrayAccessDataAsByte(SafeArrayPtr psa)
//{
//	void* p;
//	HRESULT hr = SafeArrayAccessData(psa.get(), &p);
//	if (SUCCEEDED(hr))
//	{
//		SafeArrayAccessBytePtr ret(static_cast<BYTE*>(p), SafeArrayAccessDeleter(psa.get()));
//		return boost::move(ret);
//	}
//	else
//	{
//		return SafeArrayAccessBytePtr();
//	}
//}

struct SafeArrayAccessBytePtr : boost::noncopyable
{
	explicit SafeArrayAccessBytePtr(SafeArrayPtr const& psa) : p(), psa(psa.get())
	{
		HRESULT hr = SafeArrayAccessData(this->psa, reinterpret_cast<void**>(&p));
	}

	BYTE* get() const
	{
		return p;
	}

private:
	BYTE* p;
	SAFEARRAY* psa;
};

#define TTA_VERSION L(VERSION)

#define RIFF_SIGN (0x46464952)
#define WAVE_SIGN (0x45564157)
#define fmt_SIGN  (0x20746D66)
#define data_SIGN (0x61746164)

#define PCM_BUFFER_LENGTH 5120

typedef struct
{
	TTAuint32 chunk_id;
	TTAuint32 chunk_size;
	TTAuint32 format;
	TTAuint32 subchunk_id;
	TTAuint32 subchunk_size;
	TTAuint16 audio_format;
	TTAuint16 num_channels;
	TTAuint32 sample_rate;
	TTAuint32 byte_rate;
	TTAuint16 block_align;
	TTAuint16 bits_per_sample;
} WAVE_hdr;

struct WAVE_subchunk_hdr
{
	TTAuint32 subchunk_id;
	TTAuint32 subchunk_size;
};

struct WAVE_subformat
{
	TTAuint32 f1;
	TTAuint16 f2;
	TTAuint16 f3;
	TTAuint8 f4[8];
};

struct WAVE_ext_hdr
{
	TTAuint16 cb_size;
	TTAuint16 valid_bits;
	TTAuint32 ch_mask;
	WAVE_subformat est;
};

struct TTA_io_callback_wrapper
{
	TTA_io_callback iocb;
	HANDLE handle;
};

void write_wav_hdr(std::ostream& os, WAVE_hdr const* waveHeader, TTAuint32 dataSize)
{
	WAVE_subchunk_hdr subchunk_hdr;

	subchunk_hdr.subchunk_id = data_SIGN;
	subchunk_hdr.subchunk_size = dataSize;

	// Write WAVE header
	os.write(reinterpret_cast<char const*>(waveHeader), sizeof (WAVE_hdr));

	// Write Subchunk header
	os.write(reinterpret_cast<char const*>(&subchunk_hdr), sizeof (WAVE_subchunk_hdr));
}

struct TTA_io_callback_data : TTA_io_callback
{
	BYTE const* source;
	UINT length;
	UINT position;
};

void CALLBACK tta_callback(TTAuint32 rate, TTAuint32 fnum, TTAuint32 frames)
{
	TTAuint32 pcnt = static_cast<TTAuint32>(fnum * 100. / frames);
	if ((pcnt % 10) == 0)
	{
		DebugWriteLineF(L"Progress: %02d%%", pcnt);
	}
}

TTAint32 CALLBACK read_callback(TTA_io_callback* io, TTAuint8* buffer, TTAuint32 size)
{
	TTA_io_callback_data* iocb = static_cast<TTA_io_callback_data*>(io);

	UINT unreadSize = iocb->length - iocb->position;
	if (size < unreadSize)
	{
		memcpy(buffer, iocb->source + iocb->position, size);
		iocb->position += size;
		return size;
	}
	else
	{
		memcpy(buffer, iocb->source + iocb->position, unreadSize);
		iocb->position = iocb->length;
		return unreadSize;
	}
}

TTAint64 CALLBACK seek_callback(TTA_io_callback* io, TTAint64 offset)
{
	TTA_io_callback_data* iocb = static_cast<TTA_io_callback_data*>(io);
	iocb->position = offset;
	return 0;
}

SAFEARRAY* decompress(BYTE const* source, UINT length)
{
	boost::interprocess::basic_vectorstream<std::vector<char>> os;
	os.reserve(1024 * 1024 * 2);

	TTA_io_callback_data io;
	io.read = &read_callback;
	io.seek = &seek_callback;
	io.write = NULL;

	io.source = source;
	io.length = length;
	io.position = 0;

	tta::tta_decoder tta(&io);

	TTA_info info;
	try
	{
		tta.init_get_info(&info, 0);
	}
	catch (tta::tta_exception const& e)
	{
		DebugWriteLineF(L"init_get_info: %d", e.code());
		throw;
	}

	TTAuint32 smp_size = info.nch * ((info.bps + 7) / 8);
	TTAuint32 buf_size = PCM_BUFFER_LENGTH * smp_size;

	// allocate memory for PCM buffer
	boost::scoped_array<TTAuint8> buffer(new TTAuint8[buf_size + 4]); // +4 for WRITE_BUFFER macro

	// Fill in WAV header
	TTAuint32 data_size = info.samples * smp_size;
	WAVE_hdr wave_hdr = {};
	wave_hdr.chunk_id = RIFF_SIGN;
	wave_hdr.chunk_size = data_size + 36;
	wave_hdr.format = WAVE_SIGN;
	wave_hdr.subchunk_id = fmt_SIGN;
	wave_hdr.subchunk_size = 16;
	wave_hdr.audio_format = 1;
	wave_hdr.num_channels = (TTAuint16) info.nch;
	wave_hdr.sample_rate = info.sps;
	wave_hdr.bits_per_sample = info.bps;
	wave_hdr.byte_rate = info.sps * smp_size;
	wave_hdr.block_align = static_cast<TTAuint16>(smp_size);

	// Write WAVE header
	write_wav_hdr(os, &wave_hdr, data_size);

	try
	{
		for (;;)
		{
			TTAint32 len = tta.process_stream(buffer.get(), buf_size, tta_callback);
			if (len == 0)
			{
				break;
			}
			os.write(reinterpret_cast<char const*>(buffer.get()), len * smp_size);
		}
	}
	catch (tta::tta_exception const& e)
	{
		DebugWriteLineF(L"process_stream: %d", e.code());
		throw;
	}

	std::vector<char> const& osBuffer = os.vector();
	SafeArrayPtr psaRet(SafeArrayCreateVector(VT_UI1, 0, osBuffer.size()));
	{
		SafeArrayAccessBytePtr pBuf(psaRet);
		if (pBuf.get())
		{
			memcpy(pBuf.get(), &osBuffer[0], osBuffer.size());
			return psaRet.release();
		}
	}
	return NULL;
}

class __declspec(uuid("D944B6E8-A553-4E4C-9E80-F145D5403927")) ATL_NO_VTABLE DecoderImpl :
	boost::noncopyable,
	public ATL::CComObjectRootEx<ATL::CComMultiThreadModel>,
	public ATL::CComCoClass<DecoderImpl, &__uuidof(DecoderImpl)>,
	public IDecoder
{
public:
	DecoderImpl() {}

	DECLARE_NOT_AGGREGATABLE(DecoderImpl)
	DECLARE_NO_REGISTRY()

	BEGIN_COM_MAP(DecoderImpl)
		COM_INTERFACE_ENTRY(IDecoder)
	END_COM_MAP()

	HRESULT Init(IDebug* debug) override
	{
		if (debug == NULL)
		{
			return E_INVALIDARG;
		}
		g_Debug = debug;
		return S_OK;
	}

	HRESULT Decode(BYTE* source, int length, SAFEARRAY** dest) override
	{
		DebugWriteLineF(L"DecoderImpl.Decode");

		if (dest == NULL)
		{
			return E_POINTER;
		}
		try
		{
			SAFEARRAY* ret = decompress(source, static_cast<UINT>(length));
			if (ret == NULL)
			{
				*dest = NULL;
				return E_FAIL;
			}
			else
			{
				*dest = ret;
				return S_OK;
			}
		}
		catch (std::bad_alloc const&)
		{
			return E_OUTOFMEMORY;
		}
		catch (std::exception const& e)
		{
			ATL::CA2W msg(e.what());
			DebugWriteLineF(L"DecoderImpl.Decode: %s", boost::implicit_cast<WCHAR const*>(msg));
			return E_FAIL;
		}
		catch (...)
		{
			DebugWriteLineF(L"DecoderImpl.Decode: Unknown error");
			return E_UNEXPECTED;
		}
	}
};

OBJECT_ENTRY_AUTO(__uuidof(DecoderImpl), DecoderImpl)
